'use strict';

var ethers = require('ethers');
var crossFetch = require('cross-fetch');

/******************************************************************************
Copyright (c) Microsoft Corporation.

Permission to use, copy, modify, and/or distribute this software for any
purpose with or without fee is hereby granted.

THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT,
INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
PERFORMANCE OF THIS SOFTWARE.
***************************************************************************** */
/* global Reflect, Promise, SuppressedError, Symbol */


function __awaiter(thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
}

typeof SuppressedError === "function" ? SuppressedError : function (error, suppressed, message) {
    var e = new Error(message);
    return e.name = "SuppressedError", e.error = error, e.suppressed = suppressed, e;
};

const isNode = typeof process !== 'undefined' &&
    process.versions != null &&
    process.versions.node != null;
/**
 * @see https://github.com/jsdom/jsdom/releases/tag/12.0.0
 * @see https://github.com/jsdom/jsdom/issues/1537
 */
const isJsDom = (typeof window !== 'undefined' && window.name === 'nodejs') ||
    (typeof navigator !== 'undefined' && navigator.userAgent && Array.isArray(navigator.userAgent) &&
        (navigator.userAgent.includes('Node.js') ||
            navigator.userAgent.includes('jsdom')));
const xfetch = (() => {
    if (isNode || isJsDom) {
        return crossFetch;
    }
    return fetch;
})();

const CPChainNetworkType = { name: 'CPChain', path: "m/44'/337'/0'/0/0" };

const CPChainNetwork = {
    type: CPChainNetworkType,
    name: 'CPChain',
    chainId: 337,
    nodes: [
        'https://civilian.cpchain.io'
    ],
    centralizedOpts: {
        baseUrl: 'https://api.cpchain.io'
    }
};

const TestnetNetwork = {
    type: CPChainNetworkType,
    name: 'Testnet',
    chainId: 41,
    nodes: [
        'https://civilian.testnet.cpchain.io'
    ],
    centralizedOpts: {
        baseUrl: 'https://api.testnet.cpchain.io'
    }
};

class CPCScanProviderImpl {
    constructor(networkish) {
        if (typeof networkish !== 'string') {
            this.network = networkish;
        }
        else if (networkish === 'CPChain') {
            this.network = CPChainNetwork;
        }
        else if (networkish === 'Testnet') {
            this.network = TestnetNetwork;
        }
        else {
            throw new Error(`Don't find ${networkish}`);
        }
    }
    getUrl(path) {
        const base = this.getBaseUrl();
        const lastOfBase = base[base.length - 1];
        if (!path) {
            return base;
        }
        if (lastOfBase === '/' && path[0] === '/') {
            path = path.substring(1);
        }
        else if (lastOfBase !== '/' && path[0] !== '/') {
            path = '/' + path;
        }
        return this.getBaseUrl() + path;
    }
    detectNetwork() {
        return __awaiter(this, void 0, void 0, function* () {
            return this.network;
        });
    }
    getHistory(addressOrName, props) {
        return __awaiter(this, void 0, void 0, function* () {
            const limit = ((props === null || props === void 0 ? void 0 : props.limit) || 20);
            const page = (props === null || props === void 0 ? void 0 : props.page) === undefined ? 1 : props.page;
            const params = new URLSearchParams(Object.assign({ address: addressOrName, limit: '' + limit, page: '' + page }, ((props === null || props === void 0 ? void 0 : props.others) || {})));
            const response = yield xfetch(this.getUrl('/chain/tx/?') + params);
            const res = yield response.json();
            return res;
        });
    }
    getBaseUrl() {
        return this.network.centralizedOpts.baseUrl;
    }
}
const createCPCScanProvider = ({ networkish }) => {
    return new CPCScanProviderImpl(networkish);
};

class CPCTransactionResponse {
    constructor(hash, provider, nonce, from) {
        this.hash = hash;
        this.provider = provider;
        this.confirmations = 3;
        this.from = from || '';
        this.nonce = nonce || 0;
        this.gasLimit = ethers.ethers.BigNumber.from(0);
        this.data = '';
        this.value = ethers.ethers.BigNumber.from(0);
        this.chainId = 0;
    }
    wait(confirmations) {
        return this.provider.waitForTransaction(this.hash, confirmations);
    }
}
const networks = [
    {
        name: 'Mainnet',
        chainId: 337,
        ensAddress: 'https://civilian.cpchain.io'
    }, {
        name: 'Testnet',
        chainId: 41,
        ensAddress: 'https://testnet.civilian.cpchain.io'
    }
];
function handleNumber(value) {
    if (value === '0x') {
        return ethers.ethers.constants.Zero;
    }
    return ethers.ethers.BigNumber.from(value);
}
function _parseNonce(rawTransaction) {
    const payload = ethers.ethers.utils.arrayify(rawTransaction);
    const transaction = ethers.ethers.utils.RLP.decode(payload);
    // from
    let chainId = 0;
    let fromAddr = '';
    let v = ethers.ethers.BigNumber.from(transaction[7]).toNumber();
    const r = ethers.ethers.utils.hexZeroPad(transaction[8], 32);
    const s = ethers.ethers.utils.hexZeroPad(transaction[9], 32);
    if (ethers.ethers.BigNumber.from(r).isZero() && ethers.ethers.BigNumber.from(s).isZero()) {
        // EIP-155 unsigned transaction
        chainId = v;
        v = 0;
    }
    else {
        // Signed Transaction
        chainId = Math.floor((v - 35) / 2);
        if (chainId < 0) {
            chainId = 0;
        }
        let recoveryParam = v - 27;
        const raw = transaction.slice(0, 7);
        if (chainId !== 0) {
            raw.push(ethers.ethers.utils.hexlify(chainId));
            raw.push('0x');
            raw.push('0x');
            recoveryParam -= chainId * 2 + 8;
        }
        const digest = ethers.ethers.utils.keccak256(ethers.ethers.utils.RLP.encode(raw));
        fromAddr = ethers.ethers.utils.recoverAddress(digest, { r: ethers.ethers.utils.hexlify(r), s: ethers.ethers.utils.hexlify(s), recoveryParam: recoveryParam });
    }
    return {
        nonce: handleNumber(transaction[1]).toNumber(),
        from: fromAddr
    };
}
class CPCJsonRpcProvider extends ethers.ethers.providers.JsonRpcProvider {
    constructor(url, network) {
        super(url, network);
        this._url = url;
    }
    detectNetwork() {
        return this.getNetwork();
    }
    getNetwork() {
        const network = networks.find(network => network.ensAddress === this._url) || {
            name: 'Unknown',
            chainId: this._network && this._network.chainId,
            ensAddress: this._url
        };
        return Promise.resolve(network);
    }
    sendTransaction(signedTransaction) {
        const tx = _parseNonce(signedTransaction);
        return xfetch(this._url, {
            method: 'POST',
            headers: { 'Content-type': 'application/json;charset=UTF-8' },
            body: `{"jsonrpc":"2.0","method":"eth_sendRawTransaction","params":["${signedTransaction}"],"id":1}`
        }).then((response) => __awaiter(this, void 0, void 0, function* () {
            const res = yield response.json();
            if (res.error) {
                throw new Error(res.error.message);
            }
            return (new CPCTransactionResponse((res).result, this, tx.nonce, tx.from));
        }));
    }
}
function createJsonRpcProvider(url = 'https://civilian.cpchain.io', chainID = 337) {
    return new CPCJsonRpcProvider(url, chainID);
}

exports.CPCJsonRpcProvider = CPCJsonRpcProvider;
exports.CPCScanProviderImpl = CPCScanProviderImpl;
exports.CPCTransactionResponse = CPCTransactionResponse;
exports.createCPCScanProvider = createCPCScanProvider;
exports.createJsonRpcProvider = createJsonRpcProvider;
